# cython: language_level=3
from libc.stdlib cimport strtoll, strtold
from cpython.bytes cimport PyBytes_FromStringAndSize
from cpython.unicode cimport PyUnicode_DecodeUTF8, PyUnicode_DecodeASCII
from serializor.binary.ser cimport (  # type: ignore
    UINT8_ENCODE_VALUE,
    UINT16_ENCODE_VALUE,
    UINT32_ENCODE_VALUE,
    UINT64_ENCODE_VALUE,
)

# Utils: Unpack unsigned integers
cdef inline unsigned char unpack_uint8(char* data, Py_ssize_t pos):
    """Unpack (read) `UNSIGNED` 8-bit integer from 'data'
    at the given 'pos' in little-endian order `<'int'>`.
    """
    return <unsigned char> data[pos]

cdef inline unsigned short unpack_uint16(char* data, Py_ssize_t pos):
    """Unpack (read) `UNSIGNED` 16-bit integer from 'data'
    at the given 'pos' in little-endian order `<'int'>`.
    """
    cdef:
        unsigned short p0 = <unsigned char> data[pos]
        unsigned short p1 = <unsigned char> data[pos + 1]
        unsigned short res = p0 | (p1 << 8)
    return res

cdef inline unsigned int unpack_uint24(char* data, Py_ssize_t pos):
    """Unpack (read) `UNSIGNED` 24-bit integer from 'data'
    at the given 'pos' in little-endian order `<'int'>`.
    """
    cdef:
        unsigned int p0 = <unsigned char> data[pos]
        unsigned int p1 = <unsigned char> data[pos + 1]
        unsigned int p2 = <unsigned char> data[pos + 2]
        unsigned int res = p0 | (p1 << 8) | (p2 << 16)
    return res

cdef inline unsigned int unpack_uint32(char* data, Py_ssize_t pos):
    """Unpack (read) `UNSIGNED` 32-bit integer from 'data'
    at the given 'pos' in little-endian order `<'int'>`.
    """
    cdef:
        unsigned int p0 = <unsigned char> data[pos]
        unsigned int p1 = <unsigned char> data[pos + 1]
        unsigned int p2 = <unsigned char> data[pos + 2]
        unsigned int p3 = <unsigned char> data[pos + 3]
        unsigned int res = p0 | (p1 << 8) | (p2 << 16) | (p3 << 24)
    return res

cdef inline unsigned long long unpack_uint64(char* data, Py_ssize_t pos):
    """Unpack (read) `UNSIGNED` 64-bit integer from 'data'
    at the given 'pos' in little-endian order `<'int'>`.
    """
    cdef:
        unsigned long long p0 = <unsigned char> data[pos]
        unsigned long long p1 = <unsigned char> data[pos + 1]
        unsigned long long p2 = <unsigned char> data[pos + 2]
        unsigned long long p3 = <unsigned char> data[pos + 3]
        unsigned long long p4 = <unsigned char> data[pos + 4]
        unsigned long long p5 = <unsigned char> data[pos + 5]
        unsigned long long p6 = <unsigned char> data[pos + 6]
        unsigned long long p7 = <unsigned char> data[pos + 7]
        unsigned long long res = (
            p0 | (p1 << 8) | (p2 << 16) | (p3 << 24) | (p4 << 32)
            | (p5 << 40) | (p6 << 48) | (p7 << 56) )
    return res

# Utils: Unpack signed integer
cdef inline signed char unpack_int8(char* data, Py_ssize_t pos):
    """Read (unpack) `SIGNED` 8-bit integer from 'data'
    at givent 'pos' in little-endian order `<'int'>`."""
    return <signed char> data[pos]

cdef inline short unpack_int16(char* data, Py_ssize_t pos):
    """Read (unpack) `SIGNED` 16-bit integer from 'data'
    at givent 'pos' in little-endian order `<'int'>`."""
    return <short> unpack_uint16(data, pos)

cdef inline int unpack_int24(char* data, Py_ssize_t pos):
    """Read (unpack) `SIGNED` 24-bit integer from 'data'
    at givent 'pos' in little-endian order `<'int'>`."""
    cdef int res = <int> unpack_uint24(data, pos)
    return res if res < 0x800000 else res - 0x1000000

cdef inline int unpack_int32(char* data, Py_ssize_t pos):
    """Read (unpack) `SIGNED` 32-bit integer from 'data'
    at givent 'pos' in little-endian order `<'int'>`."""
    return <int> unpack_uint32(data, pos)

cdef inline long long unpack_int64(char* data, Py_ssize_t pos):
    """Read (unpack) `SIGNED` 64-bit integer from 'data'
    at givent 'pos' in little-endian order `<'int'>`."""
    return <long long> unpack_uint64(data, pos)

# Encoded integer
cdef inline unsigned long long dec_encoded_int(char* data, Py_ssize_t pos[1]):
    """Decode encoded integer from 'data' at the given 'pos' pointer `<'int'>`.
    Note:
    - The 'pos' argument accepts the position pointer instead of the position value.
    - After decoding, the 'pos' pointer will be updated to the next position.
    """
    cdef:
        Py_ssize_t idx = pos[0]
        unsigned char code = <unsigned char> data[idx]

    if code <= UINT8_ENCODE_VALUE:
        pos[0] = idx + 1
        return code
    elif code == UINT16_ENCODE_VALUE:
        pos[0] = idx + 3
        return unpack_uint16(data, idx + 1)
    elif code == UINT32_ENCODE_VALUE:
        pos[0] = idx + 5
        return unpack_uint32(data, idx + 1)
    elif code == UINT64_ENCODE_VALUE:
        pos[0] = idx + 9
        return unpack_uint64(data, idx + 1)
    else:
        raise ValueError("Invalid encoded integer value: %d." % code)

# Slice & convert
cdef inline char* slice_to_chars(char* data, Py_ssize_t start, Py_ssize_t size):
    """Slice 'data' from 'start' with the given 'size' to `<'char*'>`."""
    cdef Py_ssize_t end = start + size
    if end < start:
        raise ValueError("Invalid slice size: [start]%d -> [end]%d." % (start, end))
    return data[start:end]

cdef inline bytes slice_to_bytes(char* data, Py_ssize_t start, Py_ssize_t size):
    """Slice 'data' from 'start' with the given 'size' to `<'bytes'>`."""
    return PyBytes_FromStringAndSize(data + start, size)

cdef inline str slice_to_str_utf8(char* data, Py_ssize_t start, Py_ssize_t size):
    """Slice 'data' from 'start' with the given 'size' to `<'str'>`.

    With 'UTF-8' encoding & 'surrogateescape' error handler.
    """
    return PyUnicode_DecodeUTF8(data + start, size, "surrogateescape")

cdef inline str slice_to_str_ascii(char* data, Py_ssize_t start, Py_ssize_t size):
    """Slice 'data' from 'start' with the given 'size' to `<'str'>`.

    With 'ascii' encoding & 'surrogateescape' error handler.
    """
    return PyUnicode_DecodeASCII(data + start, size, "surrogateescape")

cdef inline long long slice_to_int(char* data, Py_ssize_t start, Py_ssize_t size):
    """Slice 'data' from 'start' with the given 'size' to `<'int'>`."""
    return strtoll(slice_to_chars(data, start, size), NULL, 10)

cdef inline long double slice_to_float(char* data, Py_ssize_t start, Py_ssize_t size):
    """Slice 'data' from 'start' with the given 'size' to `<'float'>`."""
    return strtold(slice_to_chars(data, start, size), NULL)

# Deserialize
cpdef object deserialize(bytes data)



# cython: language_level=3
from libc.limits cimport USHRT_MAX, UINT_MAX
from cpython.bytes cimport PyBytes_FromStringAndSize

# Constants
cdef:
    # . orjson options
    object OPT_SERIALIZE_NUMPY
    # . encoded integer
    unsigned char UINT8_ENCODE_VALUE
    unsigned char UINT16_ENCODE_VALUE
    unsigned char UINT32_ENCODE_VALUE
    unsigned char UINT64_ENCODE_VALUE
    # . native types identifier
    bytes DATETIME_ID
    bytes DATE_ID
    bytes TIME_ID
    bytes TIMEDELTA_ID
    bytes STRUCT_TIME_ID
    bytes COMPLEX_ID
    bytes LIST_ID
    bytes TUPLE_ID
    bytes SET_ID
    bytes FROZENSET_ID
    bytes RANGE_ID
    bytes DICT_ID
    # . numpy identifider & dtype
    bytes DT64_ID
    bytes TD64_ID
    bytes NDARRAY_ID
    bytes NDARRAY_OBJECT_DT
    bytes NDARRAY_INT_DT
    bytes NDARRAY_UINT_DT
    bytes NDARRAY_FLOAT_DT
    bytes NDARRAY_BOOL_DT
    bytes NDARRAY_DT64_DT
    bytes NDARRAY_TD64_DT
    bytes NDARRAY_COMPLEX_DT
    bytes NDARRAY_BYTES_DT
    bytes NDARRAY_UNICODE_DT
    bytes NDARRAY_OBJECT_IDDT
    bytes NDARRAY_INT_IDDT
    bytes NDARRAY_UINT_IDDT
    bytes NDARRAY_FLOAT_IDDT
    bytes NDARRAY_BOOL_IDDT
    bytes NDARRAY_DT64_IDDT
    bytes NDARRAY_TD64_IDDT
    bytes NDARRAY_COMPLEX_IDDT
    bytes NDARRAY_BYTES_IDDT
    bytes NDARRAY_UNICODE_IDDT
    # . pandas identifider
    bytes PD_TIMESTAMP_ID
    bytes PD_TIMEDELTA_ID
    bytes SERIES_ID
    bytes DATETIMEINDEX_ID
    bytes TIMEDELTAINDEX_ID
    bytes DATAFRAME_ID
    # . value
    bytes TRUE_VALUE
    bytes FALSE_VALUE
    bytes NONE_VALUE

# Utils: Pack unsigned integers
cdef inline bytes pack_uint8(unsigned int value):
    """Pack `UNSIGNED` 8-bit integer in little-endian order to `<'bytes'>`.
    
    Equivalent to:
    >>> struct.pack("<B", value)
    """
    cdef char buffer[1]
    buffer[0] = value & 0xFF
    return PyBytes_FromStringAndSize(buffer, 1)

cdef inline bytes pack_uint16(unsigned int value):
    """Pack `UNSIGNED` 16-bit integer in little-endian order to `<'bytes'>`.
    
    Equivalent to:
    >>> struct.pack("<H", value)
    """
    cdef char buffer[2]
    buffer[0] = value & 0xFF
    buffer[1] = (value >> 8) & 0xFF
    return PyBytes_FromStringAndSize(buffer, 2)

cdef inline bytes pack_uint24(unsigned int value):
    """Pack `UNSIGNED` 24-bit integer in little-endian order to `<'bytes'>`.
    
    Equivalent to:
    >>> struct.pack("<I", value)[:3]
    """
    cdef char buffer[3]
    buffer[0] = value & 0xFF
    buffer[1] = (value >> 8) & 0xFF
    buffer[2] = (value >> 16) & 0xFF
    return PyBytes_FromStringAndSize(buffer, 3)

cdef inline bytes pack_uint32(unsigned long long value):
    """Pack `UNSIGNED` 32-bit integer in little-endian order to `<'bytes'>`.
    
    Equivalent to:
    >>> struct.pack("<I", value)
    """
    cdef char buffer[4]
    buffer[0] = value & 0xFF
    buffer[1] = (value >> 8) & 0xFF
    buffer[2] = (value >> 16) & 0xFF
    buffer[3] = (value >> 24) & 0xFF
    return PyBytes_FromStringAndSize(buffer, 4)

cdef inline bytes pack_uint64(unsigned long long value):
    """Pack `UNSIGNED` 64-bit integer in little-endian order to `<'bytes'>`.
    
    Equivalent to:
    >>> struct.pack("<Q", value)
    """
    cdef char buffer[8]
    buffer[0] = value & 0xFF
    buffer[1] = (value >> 8) & 0xFF
    buffer[2] = (value >> 16) & 0xFF
    buffer[3] = (value >> 24) & 0xFF
    buffer[4] = (value >> 32) & 0xFF
    buffer[5] = (value >> 40) & 0xFF
    buffer[6] = (value >> 48) & 0xFF
    buffer[7] = (value >> 56) & 0xFF
    return PyBytes_FromStringAndSize(buffer, 8)

# Utils: Pack signed integer
cdef inline bytes pack_int8(int value):
    """Pack `SIGNED` 8-bit integer in little-endian order to `<'bytes'>`.
    
    Equivalent to:
    >>> struct.pack("<b", value)
    """
    cdef char buffer[1]
    buffer[0] = value & 0xFF
    return PyBytes_FromStringAndSize(buffer, 1)

cdef inline bytes pack_int16(int value):
    """Pack `SIGNED` 16-bit integer in little-endian order to `<'bytes'>`.
    
    Equivalent to:
    >>> struct.pack("<h", value)
    """
    cdef char buffer[2]
    buffer[0] = value & 0xFF
    buffer[1] = (value >> 8) & 0xFF
    return PyBytes_FromStringAndSize(buffer, 2)

cdef inline bytes pack_int24(int value):
    """Pack `SIGNED` 24-bit integer in little-endian order to `<'bytes'>`.
    
    Equivalent to:
    >>> struct.pack("<i", value)[:3]
    """
    cdef char buffer[3]
    buffer[0] = value & 0xFF
    buffer[1] = (value >> 8) & 0xFF
    buffer[2] = (value >> 16) & 0xFF
    return PyBytes_FromStringAndSize(buffer, 3)

cdef inline bytes pack_int32(long long value):
    """Pack `SIGNED` 32-bit integer in little-endian order to `<'bytes'>`.
    
    Equivalent to:
    >>> struct.pack("<i", value)
    """
    cdef char buffer[4]
    buffer[0] = value & 0xFF
    buffer[1] = (value >> 8) & 0xFF
    buffer[2] = (value >> 16) & 0xFF
    buffer[3] = (value >> 24) & 0xFF
    return PyBytes_FromStringAndSize(buffer, 4)

cdef inline bytes pack_int64(long long value):
    """Pack `SIGNED` 64-bit integer in little-endian order to `<'bytes'>`.
    
    Equivalent to:
    >>> struct.pack("<q", value)
    """
    cdef char buffer[8]
    buffer[0] = value & 0xFF
    buffer[1] = (value >> 8) & 0xFF
    buffer[2] = (value >> 16) & 0xFF
    buffer[3] = (value >> 24) & 0xFF
    buffer[4] = (value >> 32) & 0xFF
    buffer[5] = (value >> 40) & 0xFF
    buffer[6] = (value >> 48) & 0xFF
    buffer[7] = (value >> 56) & 0xFF
    return PyBytes_FromStringAndSize(buffer, 8)

# Encoded integer
cdef inline bytes gen_encoded_int(unsigned long long i):
    """Generate encoded integer `<'bytes'>`."""
    cdef char buffer[9]
    if i <= UINT8_ENCODE_VALUE:
        buffer[0] = i & 0xFF
        return PyBytes_FromStringAndSize(buffer, 1)
    elif i <= USHRT_MAX:
        buffer[0] = UINT16_ENCODE_VALUE
        buffer[1] = i & 0xFF
        buffer[2] = (i >> 8) & 0xFF
        return PyBytes_FromStringAndSize(buffer, 3)
    elif i <= UINT_MAX:
        buffer[0] = UINT32_ENCODE_VALUE
        buffer[1] = i & 0xFF
        buffer[2] = (i >> 8) & 0xFF
        buffer[3] = (i >> 16) & 0xFF
        buffer[4] = (i >> 24) & 0xFF
        return PyBytes_FromStringAndSize(buffer, 5)
    else:
        buffer[0] = UINT64_ENCODE_VALUE
        buffer[1] = i & 0xFF
        buffer[2] = (i >> 8) & 0xFF
        buffer[3] = (i >> 16) & 0xFF
        buffer[4] = (i >> 24) & 0xFF
        buffer[5] = (i >> 32) & 0xFF
        buffer[6] = (i >> 40) & 0xFF
        buffer[7] = (i >> 48) & 0xFF
        buffer[8] = (i >> 56) & 0xFF
        return PyBytes_FromStringAndSize(buffer, 9)

# Header
cdef inline bytes gen_header(unsigned char identifier, unsigned long long i):
    """Generate header '[identifier]+[encoded integer]' `<'bytes'>`."""
    cdef char buffer[10]
    if i <= UINT8_ENCODE_VALUE:
        buffer[0] = identifier
        buffer[1] = i & 0xFF
        return PyBytes_FromStringAndSize(buffer, 2)
    elif i <= USHRT_MAX:
        buffer[0] = identifier
        buffer[1] = UINT16_ENCODE_VALUE
        buffer[2] = i & 0xFF
        buffer[3] = (i >> 8) & 0xFF
        return PyBytes_FromStringAndSize(buffer, 4)
    elif i <= UINT_MAX:
        buffer[0] = identifier
        buffer[1] = UINT32_ENCODE_VALUE
        buffer[2] = i & 0xFF
        buffer[3] = (i >> 8) & 0xFF
        buffer[4] = (i >> 16) & 0xFF
        buffer[5] = (i >> 24) & 0xFF
        return PyBytes_FromStringAndSize(buffer, 6)
    else:
        buffer[0] = identifier
        buffer[1] = UINT64_ENCODE_VALUE
        buffer[2] = i & 0xFF
        buffer[3] = (i >> 8) & 0xFF
        buffer[4] = (i >> 16) & 0xFF
        buffer[5] = (i >> 24) & 0xFF
        buffer[6] = (i >> 32) & 0xFF
        buffer[7] = (i >> 40) & 0xFF
        buffer[8] = (i >> 48) & 0xFF
        buffer[9] = (i >> 56) & 0xFF
        return PyBytes_FromStringAndSize(buffer, 10)

# Serialize
cpdef bytes serialize(object obj)



# cython: language_level=3
from libc.limits cimport USHRT_MAX, UINT_MAX

# Constants
cdef:
    # . orjson options
    object OPT_SERIALIZE_NUMPY
    # . encoded integer
    Py_UCS4 UINT8_ENCODE_VALUE
    Py_UCS4 UINT16_ENCODE_VALUE
    Py_UCS4 UINT32_ENCODE_VALUE
    Py_UCS4 UINT64_ENCODE_VALUE
    # . native types identifier
    str DATETIME_ID
    str DATE_ID
    str TIME_ID
    str TIMEDELTA_ID
    str STRUCT_TIME_ID
    str COMPLEX_ID
    str LIST_ID
    str TUPLE_ID
    str SET_ID
    str FROZENSET_ID
    str RANGE_ID
    str DICT_ID
    # . numpy identifider & dtype
    str DT64_ID
    str TD64_ID
    str NDARRAY_ID
    str NDARRAY_OBJECT_DT
    str NDARRAY_INT_DT
    str NDARRAY_UINT_DT
    str NDARRAY_FLOAT_DT
    str NDARRAY_BOOL_DT
    str NDARRAY_DT64_DT
    str NDARRAY_TD64_DT
    str NDARRAY_COMPLEX_DT
    str NDARRAY_BYTES_DT
    str NDARRAY_UNICODE_DT
    str NDARRAY_OBJECT_IDDT
    str NDARRAY_INT_IDDT
    str NDARRAY_UINT_IDDT
    str NDARRAY_FLOAT_IDDT
    str NDARRAY_BOOL_IDDT
    str NDARRAY_DT64_IDDT
    str NDARRAY_TD64_IDDT
    str NDARRAY_COMPLEX_IDDT
    str NDARRAY_BYTES_IDDT
    str NDARRAY_UNICODE_IDDT
    # . pandas identifider
    str PD_TIMESTAMP_ID
    str PD_TIMEDELTA_ID
    str SERIES_ID
    str DATETIMEINDEX_ID
    str TIMEDELTAINDEX_ID
    str DATAFRAME_ID
    # . value
    str TRUE_VALUE
    str FALSE_VALUE
    str NONE_VALUE

# Utils: Pack unsigned integers
cdef inline str pack_uint8(unsigned int value):
    """Pack `UNSIGNED` 8-bit integer in little-endian order to `<'str'>`.

    Equivalent to:
    >>> struct.pack("<B", value).decode("latin1")
    """
    cdef char buffer[1]
    buffer[0] = value & 0xFF
    return buffer[0:1].decode("latin1")

cdef inline str pack_uint16(unsigned int value):
    """Pack `UNSIGNED` 16-bit integer in little-endian order to `<'str'>`.

    Equivalent to:
    >>> struct.pack("<H", value).decode("latin1")
    """
    cdef char buffer[2]
    buffer[0] = value & 0xFF
    buffer[1] = (value >> 8) & 0xFF
    return buffer[0:2].decode("latin1")

cdef inline str pack_uint24(unsigned int value):
    """Pack `UNSIGNED` 24-bit integer in little-endian order to `<'str'>`.

    Equivalent to:
    >>> (struct.pack("<I", value)[:3]).decode("latin1")
    """
    cdef char buffer[3]
    buffer[0] = value & 0xFF
    buffer[1] = (value >> 8) & 0xFF
    buffer[2] = (value >> 16) & 0xFF
    return buffer[0:3].decode("latin1")

cdef inline str pack_uint32(unsigned long long value):
    """Pack `UNSIGNED` 32-bit integer in little-endian order to `<'str'>`.

    Equivalent to:
    >>> struct.pack("<I", value).decode("latin1")
    """
    cdef char buffer[4]
    buffer[0] = value & 0xFF
    buffer[1] = (value >> 8) & 0xFF
    buffer[2] = (value >> 16) & 0xFF
    buffer[3] = (value >> 24) & 0xFF
    return buffer[0:4].decode("latin1")

cdef inline str pack_uint64(unsigned long long value):
    """Pack `UNSIGNED` 64-bit integer in little-endian order to `<'str'>`.

    Equivalent to:
    >>> struct.pack("<Q", value).decode("latin1")
    """
    cdef char buffer[8]
    buffer[0] = value & 0xFF
    buffer[1] = (value >> 8) & 0xFF
    buffer[2] = (value >> 16) & 0xFF
    buffer[3] = (value >> 24) & 0xFF
    buffer[4] = (value >> 32) & 0xFF
    buffer[5] = (value >> 40) & 0xFF
    buffer[6] = (value >> 48) & 0xFF
    buffer[7] = (value >> 56) & 0xFF
    return buffer[0:8].decode("latin1")

# Utils: Pack signed integer
cdef inline str pack_int8(int value):
    """Pack `SIGNED` 8-bit integer in little-endian order to `<'str'>`.
    
    Equivalent to:
    >>> struct.pack("<b", value).encode("latin1")
    """
    cdef char buffer[1]
    buffer[0] = value & 0xFF
    return buffer[0:1].decode("latin1")

cdef inline str pack_int16(int value):
    """Pack `SIGNED` 16-bit integer in little-endian order to `<'str'>`.

    Equivalent to:
    >>> struct.pack("<h", value).decode("latin1")
    """
    cdef char buffer[2]
    buffer[0] = value & 0xFF
    buffer[1] = (value >> 8) & 0xFF
    return buffer[0:2].decode("latin1")

cdef inline str pack_int24(int value):
    """Pack `SIGNED` 24-bit integer in little-endian order to `<'str'>`.

    Equivalent to:
    >>> (struct.pack("<i", value)[:3]).decode("latin1")
    """
    cdef char buffer[3]
    buffer[0] = value & 0xFF
    buffer[1] = (value >> 8) & 0xFF
    buffer[2] = (value >> 16) & 0xFF
    return buffer[0:3].decode("latin1")

cdef inline str pack_int32(long long value):
    """Pack `SIGNED` 32-bit integer in little-endian order to `<'str'>`.

    Equivalent to:
    >>> struct.pack("<i", value).decode("latin1")
    """
    cdef char buffer[4]
    buffer[0] = value & 0xFF
    buffer[1] = (value >> 8) & 0xFF
    buffer[2] = (value >> 16) & 0xFF
    buffer[3] = (value >> 24) & 0xFF
    return buffer[0:4].decode("latin1")

cdef inline str pack_int64(long long value):
    """Pack `SIGNED` 64-bit integer in little-endian order to `<'str'>`.

    Equivalent to:
    >>> struct.pack("<q", value).decode("latin1")
    """
    cdef char buffer[8]
    buffer[0] = value & 0xFF
    buffer[1] = (value >> 8) & 0xFF
    buffer[2] = (value >> 16) & 0xFF
    buffer[3] = (value >> 24) & 0xFF
    buffer[4] = (value >> 32) & 0xFF
    buffer[5] = (value >> 40) & 0xFF
    buffer[6] = (value >> 48) & 0xFF
    buffer[7] = (value >> 56) & 0xFF
    return buffer[0:8].decode("latin1")

# Encoded integer
cdef inline str gen_encoded_int(unsigned long long i):
    """Generate encoded integer `<'str'>`."""
    cdef char buffer[9]
    if i <= UINT8_ENCODE_VALUE:
        buffer[0] = i & 0xFF
        return buffer[0:1].decode("latin1")
    elif i <= USHRT_MAX:
        buffer[0] = UINT16_ENCODE_VALUE
        buffer[1] = i & 0xFF
        buffer[2] = (i >> 8) & 0xFF
        return buffer[0:3].decode("latin1")
    elif i <= UINT_MAX:
        buffer[0] = UINT32_ENCODE_VALUE
        buffer[1] = i & 0xFF
        buffer[2] = (i >> 8) & 0xFF
        buffer[3] = (i >> 16) & 0xFF
        buffer[4] = (i >> 24) & 0xFF
        return buffer[:5].decode("latin1")
    else:
        buffer[0] = UINT64_ENCODE_VALUE
        buffer[1] = i & 0xFF
        buffer[2] = (i >> 8) & 0xFF
        buffer[3] = (i >> 16) & 0xFF
        buffer[4] = (i >> 24) & 0xFF
        buffer[5] = (i >> 32) & 0xFF
        buffer[6] = (i >> 40) & 0xFF
        buffer[7] = (i >> 48) & 0xFF
        buffer[8] = (i >> 56) & 0xFF
        return buffer[0:9].decode("latin1")

# Header
cdef inline str gen_header(unsigned char identifier, unsigned long long i):
    """Generate header '[identifier]+[encoded integer]' `<'str'>`."""
    cdef char buffer[10]
    if i <= UINT8_ENCODE_VALUE:
        buffer[0] = identifier
        buffer[1] = i & 0xFF
        return buffer[0:2].decode("latin1")
    elif i <= USHRT_MAX:
        buffer[0] = identifier
        buffer[1] = UINT16_ENCODE_VALUE
        buffer[2] = i & 0xFF
        buffer[3] = (i >> 8) & 0xFF
        return buffer[0:4].decode("latin1")
    elif i <= UINT_MAX:
        buffer[0] = identifier
        buffer[1] = UINT32_ENCODE_VALUE
        buffer[2] = i & 0xFF
        buffer[3] = (i >> 8) & 0xFF
        buffer[4] = (i >> 16) & 0xFF
        buffer[5] = (i >> 24) & 0xFF
        return buffer[:6].decode("latin1")
    else:
        buffer[0] = identifier
        buffer[1] = UINT64_ENCODE_VALUE
        buffer[2] = i & 0xFF
        buffer[3] = (i >> 8) & 0xFF
        buffer[4] = (i >> 16) & 0xFF
        buffer[5] = (i >> 24) & 0xFF
        buffer[6] = (i >> 32) & 0xFF
        buffer[7] = (i >> 40) & 0xFF
        buffer[8] = (i >> 48) & 0xFF
        buffer[9] = (i >> 56) & 0xFF
        return buffer[0:10].decode("latin1")

# Serialize
cpdef str serialize(object obj)



# cython: language_level=3
# cython: wraparound=False
# cython: boundscheck=False

# Cython imports
import cython

# Python imports
import numpy as np

# Basic Types
STR: cython.char = ord("s")
BOOL: cython.char = ord("o")
FLOAT: cython.char = ord("f")
INT: cython.char = ord("i")
NONE: cython.char = ord("n")

# Date&Time Types
DATE: cython.char = ord("d")
TIME: cython.char = ord("t")
DATETIME: cython.char = ord("z")
TIMEDELTA: cython.char = ord("l")
STRUCT_TIME: cython.char = ord("r")

# Numeric Types
DECIMAL: cython.char = ord("e")
COMPLEX: cython.char = ord("c")

# Bytes Types
BYTES: cython.char = ord("b")
BYTEARRAY: cython.char = ord("y")

# Sequence Types
LIST: cython.char = ord("L")
TUPLE: cython.char = ord("T")
SET: cython.char = ord("E")
FROZENSET: cython.char = ord("R")
RANGE: cython.char = ord("g")

# Mapping Types
DICT: cython.char = ord("D")

# NumPy Types
# fmt: off
DATETIME64: cython.char = ord("M")
TIMEDELTA64: cython.char = ord("m")
NDARRAY: cython.char = ord("N")
NDARRAY_OBJECT: cython.char = ord(np.array(None, dtype="O").dtype.kind)  # 'O'
NDARRAY_INT: cython.char = ord(np.array(1, dtype=np.int64).dtype.kind)  # 'i'
NDARRAY_UINT: cython.char = ord(np.array(1, dtype=np.uint64).dtype.kind)  # 'u'
NDARRAY_FLOAT: cython.char = ord(np.array(0.1, dtype=np.float64).dtype.kind)  # 'f'
NDARRAY_BOOL: cython.char = ord(np.array(True, dtype=bool).dtype.kind)  # 'b'
NDARRAY_DT64: cython.char = ord(np.array(1, dtype="datetime64[ns]").dtype.kind)  # 'M'
NDARRAY_TD64: cython.char = ord(np.array(1, dtype="timedelta64[ns]").dtype.kind)  # 'm'
NDARRAY_COMPLEX: cython.char = ord(np.array(1 + 1j, dtype=np.complex128).dtype.kind)  # 'c'
NDARRAY_BYTES: cython.char = ord(np.array(b"1", dtype="S").dtype.kind)  # 'S'
NDARRAY_UNICODE: cython.char = ord(np.array("1", dtype="U").dtype.kind)  # 'U'
# fmt: on

# Pandas Types
SERIES: cython.char = ord("S")
DATAFRAME: cython.char = ord("F")
PD_TIMESTAMP: cython.char = ord("p")
PD_TIMEDELTA: cython.char = ord("a")
DATETIMEINDEX: cython.char = ord("Z")
TIMEDELTAINDEX: cython.char = ord("X")


### Test for duplicates ###
_duplicates_test: cython.bint = False
if not _duplicates_test:

    def _test_duplicates(category: str, ids: list[str]) -> None:
        seen: set[str] = set()
        for i in ids:
            assert i not in seen, f"Duplicated %s: %r." % (category, i)
            seen.add(i)

    _object_ids = [
        # Basic Types
        STR,
        BOOL,
        FLOAT,
        INT,
        NONE,
        # Date&Time Types
        DATE,
        TIME,
        DATETIME,
        TIMEDELTA,
        STRUCT_TIME,
        # Numeric Types
        DECIMAL,
        COMPLEX,
        # Bytes Types
        BYTES,
        BYTEARRAY,
        # Sequence Types
        LIST,
        TUPLE,
        SET,
        FROZENSET,
        RANGE,
        # Mapping Types
        DICT,
        # Numpy Types
        DATETIME64,
        TIMEDELTA64,
        NDARRAY,
        # Pandas Types
        SERIES,
        DATAFRAME,
        PD_TIMESTAMP,
        PD_TIMEDELTA,
        DATETIMEINDEX,
        TIMEDELTAINDEX,
    ]
    _test_duplicates("object identifier", _object_ids)
    _ndarray_dtype_ids = [
        NDARRAY_OBJECT,
        NDARRAY_INT,
        NDARRAY_UINT,
        NDARRAY_FLOAT,
        NDARRAY_BOOL,
        NDARRAY_DT64,
        NDARRAY_TD64,
        NDARRAY_COMPLEX,
        NDARRAY_BYTES,
        NDARRAY_UNICODE,
    ]
    _test_duplicates("ndarray dtype identifier", _ndarray_dtype_ids)

    del _test_duplicates, _object_ids, _ndarray_dtype_ids

    _duplicates_test = True



import cython
from numpy import ndarray

# Constants
AVAILABLE_TIMEZONES: set[str]

# Utils: encode / decode
def encode_str(obj: str, encoding: cython.pchar) -> bytes:
    """(cfunc) Encode string to bytes using the 'encoding' with
    'surrogateescape' error handling `<'bytes'>`."""

def decode_bytes(data: bytes, encoding: cython.pchar) -> str:
    """(cfunc) Decode bytes to string using the 'encoding' with
    "surrogateescape" error handling `<'str'>`."""

def decode_bytes_utf8(data: bytes) -> str:
    """(cfunc) Decode bytes to string using 'utf-8' encoding with
    'surrogateescape' error handling `<'str'>`."""

def decode_bytes_ascii(data: bytes) -> str:
    """(cfunc) Decode bytes to string using 'ascii' encoding with
    'surrogateescape' error handling `<'str'>`."""

# NumPy: ndarray get item
def arr_getitem_1d(arr: ndarray, i: int) -> object:
    """(cfunc) Get item from 1-dimensional numpy ndarray as `<'object'>`."""

def arr_getitem_2d(arr: ndarray, i: int, j: int) -> object:
    """(cfunc) Get item from 2-dimensional numpy ndarray as `<'object'>`."""

def arr_getitem_3d(arr: ndarray, i: int, j: int, k: int) -> object:
    """(cfunc) Get item from 3-dimensional numpy ndarray as `<'object'>`."""

def arr_getitem_4d(arr: ndarray, i: int, j: int, k: int, l: int) -> object:
    """(cfunc) Get item from 4-dimensional numpy ndarray as `<'object'>`."""

# NumPy: ndarray set item
def arr_setitem_1d(arr: ndarray, item: object, i: int) -> bool:
    """(cfunc) Set item for 1-dimensional numpy ndarray."""

def arr_setitem_2d(arr: ndarray, item: object, i: int, j: int) -> bool:
    """(cfunc) Set item for 2-dimensional numpy ndarray."""

def arr_setitem_3d(arr: ndarray, item: object, i: int, j: int, k: int) -> bool:
    """(cfunc) Set item for 3-dimensional numpy ndarray."""

def arr_setitem_4d(arr: ndarray, item: object, i: int, j: int, k: int, l: int) -> bool:
    """(cfunc) Set item for 4-dimensional numpy ndarray."""

# NumPy: ndarray flatten
def arr_flatten(arr: ndarray) -> ndarray:
    """(cfunc) Flatten multi-dimensional numpy ndarray into
    1-dimension with 'C-ORDER' `<'np.ndarray'>`."""

# NumPy: nptime unit
def map_nptime_unit_int2str(unit: int) -> str:
    """(cfunc) Map numpy datetime64/timedelta64 unit from integer
    to the corresponding string representation `<'str'>`."""

def map_nptime_unit_str2int(unit: str) -> int:
    """(cfunc) Map numpy datetime64/timedelta64 unit from string
    representation to the corresponding integer `<'int'>`."""

def parse_arr_nptime_unit(arr: ndarray) -> int:
    """(cfunc) Parse ndarray[datetime64/timedelta64] unit from the,
    returns the unit in `<'int'>`."""



# cython: language_level=3

# Cython imports
import cython

# Python imports
import numpy as np
from decimal import Decimal
from time import struct_time
from collections import deque
from pandas import Series, DataFrame
from pandas import Timestamp, DatetimeIndex
from pandas import Timedelta, TimedeltaIndex

# Constants -------------------------------------------------------------------------
# . python types
NONE: type = type(None)
DICT_KEYS: type = type(dict().keys())
DICT_VALUES: type = type(dict().values())
DICT_ITEMS: type = type(dict().items())
DECIMAL: type[Decimal] = Decimal
STRUCT_TIME: type[struct_time] = struct_time
DEQUE: type[deque] = deque
# . numpy types
FLOAT16: type[np.float16] = np.float16
FLOAT32: type[np.float32] = np.float32
FLOAT64: type[np.float64] = np.float64
INT8: type[np.int8] = np.int8
INT16: type[np.int16] = np.int16
INT32: type[np.int32] = np.int32
INT64: type[np.int64] = np.int64
UINT8: type[np.uint8] = np.uint8
UINT16: type[np.uint16] = np.uint16
UINT32: type[np.uint32] = np.uint32
UINT64: type[np.uint64] = np.uint64
COMPLEX64: type[np.complex64] = np.complex64
COMPLEX128: type[np.complex128] = np.complex128
DATETIME64: type[np.datetime64] = np.datetime64
TIMEDELTA64: type[np.timedelta64] = np.timedelta64
STR_: type[np.str_] = np.str_
BOOL_: type[np.bool_] = np.bool_
BYTES_: type[np.bytes_] = np.bytes_
RECORD: type[np.record] = np.record
# . pandas types
SERIES: type[Series] = Series
DATAFRAME: type[DataFrame] = DataFrame
PD_TIMESTAMP: type[Timestamp] = Timestamp
PD_TIMEDELTA: type[Timedelta] = Timedelta
DATETIMEINDEX: type[DatetimeIndex] = DatetimeIndex
TIMEDELTAINDEX: type[TimedeltaIndex] = TimedeltaIndex
# . cytimes types
try:
    from cytimes import pydt, pddt

    CYTIMES_AVAILABLE: cython.bint = True
    PYDT: type[pydt] = pydt
    PDDT: type[pddt] = pddt
except ImportError:
    CYTIMES_AVAILABLE: cython.bint = False
    PYDT: type[pydt] = None
    PDDT: type[pydt] = None



# cython: language_level=3
# cython: wraparound=False
# cython: boundscheck=False

# Cython imports
import cython
from cython.cimports import numpy as np  # type: ignore
from cython.cimports.cpython import datetime  # type: ignore
from cython.cimports.cpython.list import PyList_AsTuple as list_to_tuple  # type: ignore
from cython.cimports.cpython.unicode import PyUnicode_GET_LENGTH as str_len  # type: ignore
from cython.cimports.cpython.unicode import PyUnicode_READ_CHAR as str_read  # type: ignore
from cython.cimports.cpython.complex import PyComplex_FromDoubles as gen_complex  # type: ignore
from cython.cimports.serializor import identifier, typeref, utils  # type: ignore

np.import_array()
np.import_umath()
datetime.import_datetime()

# Python imports
import numpy as np, datetime
from zoneinfo import ZoneInfo
from orjson import loads as _loads
from serializor import identifier, typeref, utils, errors

__all__ = ["deserialize"]


# Orjson loads --------------------------------------------------------------------------------
@cython.cfunc
@cython.inline(True)
def _orjson_loads(data: object) -> object:
    """(cfunc) Deserialize JSON string to python `<'object'>`.

    Based on [orjson](https://github.com/ijl/orjson) `'loads()'` function.
    """
    return _loads(data)


# Basic Types ---------------------------------------------------------------------------------
@cython.cfunc
@cython.inline(True)
def _des_str(data: str, pos: cython.Py_ssize_t[1]) -> str:
    """(cfunc) Deserialize 'data' at the given position as `<'str'>`.

    Position updates to the next charactor after deserialization.
    """
    pos[0] += 1  # skip identifier
    size: cython.ulonglong = dec_encoded_int(data, pos)  # type: ignore
    idx: cython.Py_ssize_t = pos[0]
    end: cython.Py_ssize_t = idx + size
    pos[0] = end  # set new position
    return data[idx:end]


@cython.cfunc
@cython.inline(True)
def _des_int(data: str, pos: cython.Py_ssize_t[1]) -> object:
    """(cfunc) Deserialize 'data' at the given position as `<'int'>`.

    Position updates to the next charactor after deserialization.
    """
    return int(_des_str(data, pos))


@cython.cfunc
@cython.inline(True)
def _des_float(data: str, pos: cython.Py_ssize_t[1]) -> object:
    """(cfunc) Deserialize 'data' at the given position as `<'float'>`.

    Position updates to the next charactor after deserialization.
    """
    return float(_des_str(data, pos))


@cython.cfunc
@cython.inline(True)
def _des_bool(data: str, pos: cython.Py_ssize_t[1]) -> object:
    """(cfunc) Deserialize 'data' at the given position as `<'bool'>`.

    Position updates to the next charactor after deserialization.
    """
    val: cython.uchar = str_read(data, pos[0] + 1)
    pos[0] += 2  # skip identifier & boolean value
    return True if val == 1 else False


@cython.cfunc
@cython.inline(True)
def _des_none(data: str, pos: cython.Py_ssize_t[1]) -> object:
    """(cfunc) Deserialize 'data' at the given position as `<'None'>`.

    Position updates to the next charactor after deserialization.
    """
    pos[0] += 1  # skip identifier
    return None


# Date&Time Types -----------------------------------------------------------------------------
@cython.cfunc
@cython.inline(True)
def _des_datetime(data: str, pos: cython.Py_ssize_t[1]) -> object:
    """(cfunc) Deserialize 'data' at the given position as `<'datetime.datetime'>`.

    Position updates to the next charactor after deserialization.
    """
    idx: cython.Py_ssize_t = pos[0]
    flag: cython.uchar = unpack_uint8(data, idx + 1)  # type: ignore
    yy: cython.int = unpack_uint16(data, idx + 2)  # type: ignore
    mm: cython.int = unpack_uint8(data, idx + 4)  # type: ignore
    dd: cython.int = unpack_uint8(data, idx + 5)  # type: ignore
    hh: cython.int = unpack_uint8(data, idx + 6)  # type: ignore
    mi: cython.int = unpack_uint8(data, idx + 7)  # type: ignore
    ss: cython.int = unpack_uint8(data, idx + 8)  # type: ignore
    us: cython.int = unpack_uint32(data, idx + 9)  # type: ignore
    # Without timezone
    if flag == 0:
        pos[0] += 13  # set new position
        return datetime.datetime_new(yy, mm, dd, hh, mi, ss, us, None, 0)
    # Timezone: approach 1 (tz_name)
    elif flag == 1:
        pos[0] += 13  # set new position
        size: cython.ulonglong = dec_encoded_int(data, pos)  # type: ignore
        idx = pos[0]
        end: cython.Py_ssize_t = idx + size
        tz_name: str = data[idx:end]
        pos[0] = end  # set new position
        return datetime.datetime_new(yy, mm, dd, hh, mi, ss, us, ZoneInfo(tz_name), 0)
    # Timezone: approach 2 (tz_offset)
    else:
        offset = datetime.timedelta_new(
            unpack_int32(data, idx + 13),  # type: ignore
            unpack_int32(data, idx + 17),  # type: ignore
            unpack_int32(data, idx + 21),  # type: ignore
        )
        pos[0] += 25  # set new position
        tzinfo = datetime.timezone_new(offset, None)
        return datetime.datetime_new(yy, mm, dd, hh, mi, ss, us, tzinfo, 0)


@cython.cfunc
@cython.inline(True)
def _des_date(data: str, pos: cython.Py_ssize_t[1]) -> object:
    """(cfunc) Deserialize 'data' at the given position as `<'datetime.date'>`.

    Position updates to the next charactor after deserialization.
    """
    idx: cython.Py_ssize_t = pos[0] + 1  # skip identifier
    year: cython.int = unpack_uint16(data, idx)  # type: ignore
    month: cython.int = unpack_uint8(data, idx + 2)  # type: ignore
    day: cython.int = unpack_uint8(data, idx + 3)  # type: ignore
    pos[0] += 5  # set new position
    return datetime.date_new(year, month, day)


@cython.cfunc
@cython.inline(True)
def _des_time(data: str, pos: cython.Py_ssize_t[1]) -> object:
    """(cfunc) Deserialize 'data' at the given position as `<'datetime.time'>`.

    Position updates to the next charactor after deserialization.
    """
    idx: cython.Py_ssize_t = pos[0]
    flag: cython.uchar = unpack_uint8(data, idx + 1)  # type: ignore
    hh: cython.int = unpack_uint8(data, idx + 2)  # type: ignore
    mi: cython.int = unpack_uint8(data, idx + 3)  # type: ignore
    ss: cython.int = unpack_uint8(data, idx + 4)  # type: ignore
    us: cython.int = unpack_uint32(data, idx + 5)  # type: ignore
    # Without timezone
    if flag == 0:
        pos[0] += 9
        return datetime.time_new(hh, mi, ss, us, None, 0)
    # Timezone: approach 1 (tz_name)
    elif flag == 1:
        pos[0] += 9  # set new position
        size: cython.ulonglong = dec_encoded_int(data, pos)  # type: ignore
        idx = pos[0]
        end: cython.Py_ssize_t = idx + size
        tz_name: str = data[idx:end]
        pos[0] = end  # set new position
        return datetime.time_new(hh, mi, ss, us, ZoneInfo(tz_name), 0)
    # Timezone: approach 2 (tz_offset)
    else:
        offset = datetime.timedelta_new(
            unpack_int32(data, idx + 9),  # type: ignore
            unpack_int32(data, idx + 13),  # type: ignore
            unpack_int32(data, idx + 17),  # type: ignore
        )
        pos[0] += 21  # set new position
        tzinfo = datetime.timezone_new(offset, None)
        return datetime.time_new(hh, mi, ss, us, tzinfo, 0)


@cython.cfunc
@cython.inline(True)
def _des_timedelta(data: str, pos: cython.Py_ssize_t[1]) -> object:
    """(cfunc) Deserialize 'data' at the given position as `<'datetime.timedelta'>`.

    Position updates to the next charactor after deserialization.
    """
    idx: cython.Py_ssize_t = pos[0] + 1  # skip identifier
    days: cython.int = unpack_int32(data, idx)  # type: ignore
    seconds: cython.int = unpack_int32(data, idx + 4)  # type: ignore
    microseconds: cython.int = unpack_int32(data, idx + 8)  # type: ignore
    pos[0] += 13  # set new position
    return datetime.timedelta_new(days, seconds, microseconds)


@cython.cfunc
@cython.inline(True)
def _des_struct_time(data: str, pos: cython.Py_ssize_t[1]) -> object:
    """(cfunc) Deserialize 'data' at the given position as `<'time.struct_time'>`.

    Position updates to the next charactor after deserialization.
    """
    pos[0] += 1  # skip identifier
    return typeref.STRUCT_TIME(_des_list(data, pos))


# Numeric Types -------------------------------------------------------------------------------
@cython.cfunc
@cython.inline(True)
def _des_decimal(data: str, pos: cython.Py_ssize_t[1]) -> object:
    """(cfunc) Deserialize 'data' at the given position as `<'decimal.Decimal'>`.

    Position updates to the next charactor after deserialization.
    """
    pos[0] += 1  # skip identifier
    size: cython.ulonglong = dec_encoded_int(data, pos)  # type: ignore
    idx: cython.Py_ssize_t = pos[0]
    end: cython.Py_ssize_t = idx + size
    pos[0] = end  # set new position
    return typeref.DECIMAL(data[idx:end])


@cython.cfunc
@cython.inline(True)
def _des_complex(data: str, pos: cython.Py_ssize_t[1]) -> object:
    """(cfunc) Deserialize 'data' at the given position as `<'complex'>`.

    Position updates to the next charactor after deserialization.
    """
    pos[0] += 1  # skip identifier
    # Complex: real
    size: cython.ulonglong = dec_encoded_int(data, pos)  # type: ignore
    idx: cython.Py_ssize_t = pos[0]
    end: cython.Py_ssize_t = idx + size
    real: cython.longdouble = float(data[idx:end])
    pos[0] = end  # set new position
    # Complex: imag
    size = dec_encoded_int(data, pos)  # type: ignore
    idx: cython.Py_ssize_t = pos[0]
    end: cython.Py_ssize_t = idx + size
    imag: cython.longdouble = float(data[idx:end])
    pos[0] = end  # set new position
    # Generate complex
    return gen_complex(real, imag)


# Bytes Types ---------------------------------------------------------------------------------
@cython.cfunc
@cython.inline(True)
def _des_bytes(data: str, pos: cython.Py_ssize_t[1]) -> object:
    """(cfunc) Deserialize 'data' at the given position as `<'bytes'>`.

    Position updates to the next charactor after deserialization.
    """
    pos[0] += 1  # skip identifier
    size: cython.ulonglong = dec_encoded_int(data, pos)  # type: ignore
    idx: cython.Py_ssize_t = pos[0]
    end: cython.Py_ssize_t = idx + size
    pos[0] = end  # set new position
    return utils.encode_str(data[idx:end], b"utf-8")


@cython.cfunc
@cython.inline(True)
def _des_bytearray(data: str, pos: cython.Py_ssize_t[1]) -> object:
    """(cfunc) Deserialize 'data' at the given position as `<'bytearray'>`.

    Position updates to the next charactor after deserialization.
    """
    return bytearray(_des_bytes(data, pos))


# Sequence Types ------------------------------------------------------------------------------
@cython.cfunc
@cython.inline(True)
def _des_list(data: str, pos: cython.Py_ssize_t[1]) -> list:
    """(cfunc) Deserialize 'data' at the given position as `<'list'>`.

    Position updates to the next charactor after deserialization.
    """
    pos[0] += 1  # skip identifier
    size: cython.ulonglong = dec_encoded_int(data, pos)  # type: ignore
    return [_des_data(data, pos) for _ in range(size)]


@cython.cfunc
@cython.inline(True)
def _des_tuple(data: str, pos: cython.Py_ssize_t[1]) -> object:
    """(cfunc) Deserialize 'data' at the given position as `<'tuple'>`.

    Position updates to the next charactor after deserialization.
    """
    return list_to_tuple(_des_list(data, pos))


@cython.cfunc
@cython.inline(True)
def _des_set(data: str, pos: cython.Py_ssize_t[1]) -> object:
    """(cfunc) Deserialize 'data' at the given position as `<'set'>`.

    Position updates to the next charactor after deserialization.
    """
    pos[0] += 1  # skip identifier
    size: cython.ulonglong = dec_encoded_int(data, pos)  # type: ignore
    return {_des_data(data, pos) for _ in range(size)}


@cython.cfunc
@cython.inline(True)
def _des_frozenset(data: str, pos: cython.Py_ssize_t[1]) -> object:
    """(cfunc) Deserialize 'data' at the given position as `<'frozenset'>`.

    Position updates to the next charactor after deserialization.
    """
    pos[0] += 1  # skip identifier
    size: cython.ulonglong = dec_encoded_int(data, pos)  # type: ignore
    return frozenset([_des_data(data, pos) for _ in range(size)])


@cython.cfunc
@cython.inline(True)
def _des_range(data: str, pos: cython.Py_ssize_t[1]) -> object:
    """(cfunc) Deserialize 'data' at the given position as `<'range'>`.

    Position updates to the next charactor after deserialization.
    """
    pos[0] += 1  # skip identifier
    return range(
        _des_int(data, pos),  # start
        _des_int(data, pos),  # stop
        _des_int(data, pos),  # step
    )


# Mapping Types -------------------------------------------------------------------------------
@cython.cfunc
@cython.inline(True)
def _des_dict(data: str, pos: cython.Py_ssize_t[1]) -> object:
    """(cfunc) Deserialize 'data' at the given position as `<'dict'>`.

    Position updates to the next charactor after deserialization.
    """
    pos[0] += 1  # skip identifier
    size: cython.ulonglong = dec_encoded_int(data, pos)  # type: ignore
    return {_des_data(data, pos): _des_data(data, pos) for _ in range(size)}


# NumPy Types ---------------------------------------------------------------------------------
@cython.cfunc
@cython.inline(True)
def _des_datetime64(data: str, pos: cython.Py_ssize_t[1]) -> object:
    """(cfunc) Deserialize 'data' at the given position as `<'np.datetime64'>`.

    Position updates to the next charactor after deserialization.
    """
    idx: cython.Py_ssize_t = pos[0] + 1  # skip identifier
    unit: cython.uchar = unpack_uint8(data, idx)  # type: ignore
    val: cython.longlong = unpack_int64(data, idx + 1)  # type: ignore
    pos[0] += 10  # set new position
    return typeref.DATETIME64(val, utils.map_nptime_unit_int2str(unit))


@cython.cfunc
@cython.inline(True)
def _des_timedelta64(data: str, pos: cython.Py_ssize_t[1]) -> object:
    """(cfunc) Deserialize 'data' at the given position as `<'np.timedelta64'>`.

    Position updates to the next charactor after deserialization.
    """
    idx: cython.Py_ssize_t = pos[0] + 1  # skip identifier
    unit: cython.uchar = unpack_uint8(data, idx)  # type: ignore
    val: cython.longlong = unpack_int64(data, idx + 1)  # type: ignore
    pos[0] += 10  # set new position
    return typeref.TIMEDELTA64(val, utils.map_nptime_unit_int2str(unit))


@cython.cfunc
@cython.inline(True)
def _des_ndarray(data: str, pos: cython.Py_ssize_t[1]) -> object:
    """(cfunc) Deserialize 'data' at the given position as `<'np.ndarray'>`.

    Position updates to the next charactor after deserialization.
    """
    pos[0] += 1  # skip identifier
    # Get ndarray dtype
    arr_dtype: cython.char = str_read(data, pos[0])

    # Deserialize ndarray
    # . ndarray[object]
    if arr_dtype == identifier.NDARRAY_OBJECT:
        return _des_ndarray_object(data, pos)
    # . ndarray[int]
    if arr_dtype == identifier.NDARRAY_INT:
        return _des_ndarray_numeric(data, pos)
    # . ndarray[uint]
    if arr_dtype == identifier.NDARRAY_UINT:
        return _des_ndarray_numeric(data, pos)
    # . ndarray[float]
    if arr_dtype == identifier.NDARRAY_FLOAT:
        return _des_ndarray_numeric(data, pos)
    # . ndarray[bool]
    if arr_dtype == identifier.NDARRAY_BOOL:
        return _des_ndarray_numeric(data, pos)
    # . ndarray[datetime64]
    if arr_dtype == identifier.NDARRAY_DT64:
        return _des_ndarray_nptime(data, pos, True)
    # . ndarray[timedelta64]
    if arr_dtype == identifier.NDARRAY_TD64:
        return _des_ndarray_nptime(data, pos, False)
    # . ndarray[complex]
    if arr_dtype == identifier.NDARRAY_COMPLEX:
        return _des_ndarray_complex(data, pos)
    # . ndarray[bytes]
    if arr_dtype == identifier.NDARRAY_BYTES:
        return _des_ndarray_bytes(data, pos)
    # . ndarray[str]
    if arr_dtype == identifier.NDARRAY_UNICODE:
        return _des_ndarray_unicode(data, pos)
    # . invalid dtype
    raise ValueError("unsupported <'numpy.ndarray'> dtype '%s'." % chr(arr_dtype))


@cython.cfunc
@cython.inline(True)
def _des_ndarray_object(data: str, pos: cython.Py_ssize_t[1]) -> object:
    """(cfunc) Deserialize 'data' at the given position as `<'np.ndarray'>`.

    #### This function is only for ndarray dtype: `"O" (object)`.

    Position updates to the next charactor after deserialization.
    """
    idx: cython.Py_ssize_t = pos[0] + 1  # skip array dtype
    npy_dtype = unpack_uint8(data, idx)  # type: ignore
    ndim: cython.uchar = unpack_uint8(data, idx + 1)  # type: ignore
    pos[0] += 3  # set new position
    arr: np.ndarray
    # 1-dimension
    if ndim == 1:
        # . array shape
        s_i: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        arr = np.PyArray_EMPTY(ndim, [s_i], npy_dtype, 0)
        if s_i == 0:
            return arr  # exit: empty
        # . deserialize values
        for i in range(s_i):
            utils.arr_setitem_1d(arr, _des_data(data, pos), i)
        return arr

    # 2-dimensions
    if ndim == 2:
        # . array shape
        s_i: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        s_j: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        arr = np.PyArray_EMPTY(ndim, [s_i, s_j], npy_dtype, 0)
        if s_j == 0:
            return arr  # exit: empty
        # . deserialize values
        for i in range(s_i):
            for j in range(s_j):
                utils.arr_setitem_2d(arr, _des_data(data, pos), i, j)
        return arr

    # 3-dimensions
    if ndim == 3:
        # . array shape
        s_i: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        s_j: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        s_k: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        arr = np.PyArray_EMPTY(ndim, [s_i, s_j, s_k], npy_dtype, 0)
        if s_k == 0:
            return arr  # exit: empty
        # . deserialize values
        for i in range(s_i):
            for j in range(s_j):
                for k in range(s_k):
                    utils.arr_setitem_3d(arr, _des_data(data, pos), i, j, k)
        return arr

    # 4-dimensions
    if ndim == 4:
        # . array shape
        s_i: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        s_j: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        s_k: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        s_l: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        arr = np.PyArray_EMPTY(ndim, [s_i, s_j, s_k, s_l], npy_dtype, 0)
        if s_l == 0:
            return arr  # exit: empty
        # . deserialize values
        for i in range(s_i):
            for j in range(s_j):
                for k in range(s_k):
                    for l in range(s_l):
                        utils.arr_setitem_4d(arr, _des_data(data, pos), i, j, k, l)
        return arr

    # Invalid
    raise ValueError("unsupported <'numpy.ndarray'> dimension [%d]." % ndim)


@cython.cfunc
@cython.inline(True)
def _des_ndarray_numeric(data: str, pos: cython.Py_ssize_t[1]) -> object:
    """(cfunc) Deserialize 'data' at the given position as `<'np.ndarray'>`.

    #### This function is for ndarray dtype: `"i" (int)`, `"u" (uint)`, `"f" (float)` and `"b" (bool)`.

    Position updates to the next charactor after deserialization.
    """
    idx: cython.Py_ssize_t = pos[0] + 1  # skip array dtype
    npy_dtype = unpack_uint8(data, idx)  # type: ignore
    ndim: cython.uchar = unpack_uint8(data, idx + 1)  # type: ignore
    pos[0] += 3  # set new position
    arr: np.ndarray
    # 1-dimension
    if ndim == 1:
        # . array shape
        s_i: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        arr = np.PyArray_EMPTY(ndim, [s_i], npy_dtype, 0)
        if s_i == 0:
            return arr  # exit: empty
        # . deserialize values
        values_len: cython.ulonglong = dec_encoded_int(data, pos)  # type: ignore
        idx: cython.Py_ssize_t = pos[0]
        end: cython.Py_ssize_t = idx + values_len
        values_json: str = data[idx:end]
        pos[0] = end  # set new position
        values = iter(_orjson_loads(values_json))
        for i in range(s_i):
            utils.arr_setitem_1d(arr, next(values), i)
        return arr

    # 2-dimension
    if ndim == 2:
        # . array shape
        s_i: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        s_j: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        arr = np.PyArray_EMPTY(ndim, [s_i, s_j], npy_dtype, 0)
        if s_j == 0:
            return arr  # exit: empty
        # . deserialize values
        values_len: cython.ulonglong = dec_encoded_int(data, pos)  # type: ignore
        idx: cython.Py_ssize_t = pos[0]
        end: cython.Py_ssize_t = idx + values_len
        values_json: str = data[idx:end]
        pos[0] = end  # set new position
        values = iter(_orjson_loads(values_json))
        for i in range(s_i):
            for j in range(s_j):
                utils.arr_setitem_2d(arr, next(values), i, j)
        return arr

    # 3-dimension
    if ndim == 3:
        # . array shape
        s_i: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        s_j: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        s_k: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        arr = np.PyArray_EMPTY(ndim, [s_i, s_j, s_k], npy_dtype, 0)
        if s_k == 0:
            return arr  # exit: empty
        # . deserialize values
        values_len: cython.ulonglong = dec_encoded_int(data, pos)  # type: ignore
        idx: cython.Py_ssize_t = pos[0]
        end: cython.Py_ssize_t = idx + values_len
        values_json: str = data[idx:end]
        pos[0] = end  # set new position
        values = iter(_orjson_loads(values_json))
        for i in range(s_i):
            for j in range(s_j):
                for k in range(s_k):
                    utils.arr_setitem_3d(arr, next(values), i, j, k)
        return arr

    # 4-dimension
    if ndim == 4:
        # . array shape
        s_i: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        s_j: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        s_k: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        s_l: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        arr = np.PyArray_EMPTY(ndim, [s_i, s_j, s_k, s_l], npy_dtype, 0)
        if s_l == 0:
            return arr
        # . deserialize values
        values_len: cython.ulonglong = dec_encoded_int(data, pos)  # type: ignore
        idx: cython.Py_ssize_t = pos[0]
        end: cython.Py_ssize_t = idx + values_len
        values_json: str = data[idx:end]
        pos[0] = end  # set new position
        values = iter(_orjson_loads(values_json))
        for i in range(s_i):
            for j in range(s_j):
                for k in range(s_k):
                    for l in range(s_l):
                        utils.arr_setitem_4d(arr, next(values), i, j, k, l)
        return arr

    # Invalid
    raise ValueError("unsupported <'numpy.ndarray'> dimension [%d]." % ndim)


@cython.cfunc
@cython.inline(True)
def _des_ndarray_nptime(
    data: str,
    pos: cython.Py_ssize_t[1],
    is_dt64: cython.bint,
) -> object:
    """(cfunc) Deserialize 'data' at the given position as `<'np.ndarray'>`.

    #### This function is for ndarray dtype: `"M" (datetime64)`, `"m" (timedelta64)`.

    Position updates to the next charactor after deserialization.
    """
    ndim: cython.uchar = unpack_uint8(data, pos[0] + 2)  # type: ignore
    pos[0] += 3  # set new position
    arr: np.ndarray
    # 1-dimension
    if ndim == 1:
        # . array shape
        s_i: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        unit: str = utils.map_nptime_unit_int2str(unpack_uint8(data, pos[0]))  # type: ignore
        pos[0] += 1  # skip unit
        arr = np.empty(
            s_i,
            "datetime64[%s]" % unit if is_dt64 else "timedelta64[%s]" % unit,
        )
        if s_i == 0:
            return arr  # exit: empty
        # . deserialize values
        values_len: cython.ulonglong = dec_encoded_int(data, pos)  # type: ignore
        idx: cython.Py_ssize_t = pos[0]
        end: cython.Py_ssize_t = idx + values_len
        values_json: str = data[idx:end]
        pos[0] = end  # set new position
        values = iter(_orjson_loads(values_json))
        for i in range(s_i):
            utils.arr_setitem_1d(arr, next(values), i)
        return arr

    # 2-dimension
    if ndim == 2:
        # . array shape
        s_i: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        s_j: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        unit: str = utils.map_nptime_unit_int2str(unpack_uint8(data, pos[0]))  # type: ignore
        pos[0] += 1  # skip unit
        arr = np.empty(
            [s_i, s_j],
            "datetime64[%s]" % unit if is_dt64 else "timedelta64[%s]" % unit,
        )
        if s_j == 0:
            return arr  # exit: empty
        # . deserialize values
        values_len: cython.ulonglong = dec_encoded_int(data, pos)  # type: ignore
        idx: cython.Py_ssize_t = pos[0]
        end: cython.Py_ssize_t = idx + values_len
        values_json: str = data[idx:end]
        pos[0] = end  # set new position
        values = iter(_orjson_loads(values_json))
        for i in range(s_i):
            for j in range(s_j):
                utils.arr_setitem_2d(arr, next(values), i, j)
        return arr

    # 3-dimension
    if ndim == 3:
        # . array shape
        s_i: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        s_j: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        s_k: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        unit: str = utils.map_nptime_unit_int2str(unpack_uint8(data, pos[0]))  # type: ignore
        pos[0] += 1  # skip unit
        arr = np.empty(
            [s_i, s_j, s_k],
            "datetime64[%s]" % unit if is_dt64 else "timedelta64[%s]" % unit,
        )
        if s_k == 0:
            return arr  # exit: empty
        # . deserialize values
        values_len: cython.ulonglong = dec_encoded_int(data, pos)  # type: ignore
        idx: cython.Py_ssize_t = pos[0]
        end: cython.Py_ssize_t = idx + values_len
        values_json: str = data[idx:end]
        pos[0] = end  # set new position
        values = iter(_orjson_loads(values_json))
        for i in range(s_i):
            for j in range(s_j):
                for k in range(s_k):
                    utils.arr_setitem_3d(arr, next(values), i, j, k)
        return arr

    # 4-dimension
    if ndim == 4:
        # . array shape
        s_i: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        s_j: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        s_k: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        s_l: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        unit: str = utils.map_nptime_unit_int2str(unpack_uint8(data, pos[0]))  # type: ignore
        pos[0] += 1  # skip unit
        arr = np.empty(
            [s_i, s_j, s_k, s_l],
            "datetime64[%s]" % unit if is_dt64 else "timedelta64[%s]" % unit,
        )
        if s_l == 0:
            return arr  # exit: empty
        # . deserialize values
        values_len: cython.ulonglong = dec_encoded_int(data, pos)  # type: ignore
        idx: cython.Py_ssize_t = pos[0]
        end: cython.Py_ssize_t = idx + values_len
        values_json: str = data[idx:end]
        pos[0] = end  # set new position
        values = iter(_orjson_loads(values_json))
        for i in range(s_i):
            for j in range(s_j):
                for k in range(s_k):
                    for l in range(s_l):
                        utils.arr_setitem_4d(arr, next(values), i, j, k, l)
        return arr

    # Invalid
    raise ValueError("unsupported <'numpy.ndarray'> dimension [%d]." % ndim)


@cython.cfunc
@cython.inline(True)
def _des_ndarray_complex(data: str, pos: cython.Py_ssize_t[1]) -> object:
    """(cfunc) Deserialize 'data' at the given position as `<'np.ndarray'>`.

    #### This function is for only ndarray dtype: `"c" (complex)`.

    Position updates to the next charactor after deserialization.
    """
    idx: cython.Py_ssize_t = pos[0] + 1  # skip array dtype
    npy_dtype = unpack_uint8(data, idx)  # type: ignore
    ndim: cython.uchar = unpack_uint8(data, idx + 1)  # type: ignore
    pos[0] += 3  # set new position
    arr: np.ndarray
    # 1-dimension
    if ndim == 1:
        # . array shape
        s_i: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        arr = np.PyArray_EMPTY(ndim, [s_i], npy_dtype, 0)
        if s_i == 0:
            return arr  # exit: empty
        # . deserialize values
        values_len: cython.ulonglong = dec_encoded_int(data, pos)  # type: ignore
        idx: cython.Py_ssize_t = pos[0]
        end: cython.Py_ssize_t = idx + values_len
        values_json: str = data[idx:end]
        pos[0] = end  # set new position
        values = iter(_orjson_loads(values_json))
        for i in range(s_i):
            utils.arr_setitem_1d(arr, gen_complex(next(values), next(values)), i)
        return arr

    # 2-dimension
    if ndim == 2:
        # . array shape
        s_i: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        s_j: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        arr = np.PyArray_EMPTY(ndim, [s_i, s_j], npy_dtype, 0)
        if s_j == 0:
            return arr  # exit: empty
        # . deserialize values
        values_len: cython.ulonglong = dec_encoded_int(data, pos)  # type: ignore
        idx: cython.Py_ssize_t = pos[0]
        end: cython.Py_ssize_t = idx + values_len
        values_json: str = data[idx:end]
        pos[0] = end  # set new position
        values = iter(_orjson_loads(values_json))
        for i in range(s_i):
            for j in range(s_j):
                utils.arr_setitem_2d(arr, gen_complex(next(values), next(values)), i, j)
        return arr

    # 3-dimension
    if ndim == 3:
        # . array shape
        s_i: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        s_j: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        s_k: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        arr = np.PyArray_EMPTY(ndim, [s_i, s_j, s_k], npy_dtype, 0)
        if s_k == 0:
            return arr  # exit: empty
        # . deserialize values
        values_len: cython.ulonglong = dec_encoded_int(data, pos)  # type: ignore
        idx: cython.Py_ssize_t = pos[0]
        end: cython.Py_ssize_t = idx + values_len
        values_json: str = data[idx:end]
        pos[0] = end  # set new position
        values = iter(_orjson_loads(values_json))
        for i in range(s_i):
            for j in range(s_j):
                for k in range(s_k):
                    utils.arr_setitem_3d(arr, gen_complex(next(values), next(values)), i, j, k)  # type: ignore
        return arr

    # 4-dimension
    if ndim == 4:
        # . array shape
        s_i: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        s_j: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        s_k: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        s_l: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        arr = np.PyArray_EMPTY(ndim, [s_i, s_j, s_k, s_l], npy_dtype, 0)  # type: ignore
        if s_l == 0:
            return arr
        # . deserialize values
        values_len: cython.ulonglong = dec_encoded_int(data, pos)  # type: ignore
        idx: cython.Py_ssize_t = pos[0]
        end: cython.Py_ssize_t = idx + values_len
        values_json: str = data[idx:end]
        pos[0] = end  # set new position
        values = iter(_orjson_loads(values_json))
        for i in range(s_i):
            for j in range(s_j):
                for k in range(s_k):
                    for l in range(s_l):
                        utils.arr_setitem_4d(arr, gen_complex(next(values), next(values)), i, j, k, l)  # type: ignore
        return arr

    # Invalid
    raise ValueError("unsupported <'numpy.ndarray'> dimension [%d]." % ndim)


@cython.cfunc
@cython.inline(True)
def _des_ndarray_bytes(data: str, pos: cython.Py_ssize_t[1]) -> object:
    """(cfunc) Deserialize 'data' at the given position as `<'np.ndarray'>`.

    #### This function is only for ndarray dtype: `"S" (String)`.

    Position updates to the next charactor after deserialization.
    """
    idx: cython.Py_ssize_t = pos[0] + 1  # skip array dtype
    npy_dtype = unpack_uint8(data, idx)  # type: ignore
    ndim: cython.uchar = unpack_uint8(data, idx + 1)  # type: ignore
    pos[0] += 3  # set new position
    arr: np.ndarray
    # 1-dimension
    if ndim == 1:
        # . array shape
        s_i: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        if s_i == 0:
            return np.PyArray_EMPTY(ndim, [s_i], npy_dtype, 0)  # exit: empty
        # . deserialize values
        ch_size: cython.int = dec_encoded_int(data, pos)  # type: ignore
        arr = np.empty(s_i, "S%d" % ch_size)
        for i in range(s_i):
            utils.arr_setitem_1d(arr, _des_bytes(data, pos), i)
        return arr

    # 2-dimension
    if ndim == 2:
        # . array shape
        s_i: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        s_j: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        if s_j == 0:
            return np.PyArray_EMPTY(ndim, [s_i, s_j], npy_dtype, 0)  # exit: empty
        # . deserialize values
        ch_size: cython.int = dec_encoded_int(data, pos)  # type: ignore
        arr = np.empty([s_i, s_j], "S%d" % ch_size)
        for i in range(s_i):
            for j in range(s_j):
                utils.arr_setitem_2d(arr, _des_bytes(data, pos), i, j)
        return arr

    # 3-dimension
    if ndim == 3:
        # . array shape
        s_i: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        s_j: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        s_k: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        if s_k == 0:
            return np.PyArray_EMPTY(ndim, [s_i, s_j, s_k], npy_dtype, 0)  # exit: empty
        # . deserialize values
        ch_size: cython.int = dec_encoded_int(data, pos)  # type: ignore
        arr = np.empty([s_i, s_j, s_k], "S%d" % ch_size)
        for i in range(s_i):
            for j in range(s_j):
                for k in range(s_k):
                    utils.arr_setitem_3d(arr, _des_bytes(data, pos), i, j, k)
        return arr

    # 4-dimension
    if ndim == 4:
        # . array shape
        s_i: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        s_j: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        s_k: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        s_l: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        if s_l == 0:
            return np.PyArray_EMPTY(
                ndim, [s_i, s_j, s_k, s_l], npy_dtype, 0
            )  # exit: empty
        # . deserialize values
        ch_size: cython.int = dec_encoded_int(data, pos)  # type: ignore
        arr = np.empty([s_i, s_j, s_k, s_l], "S%d" % ch_size)
        for i in range(s_i):
            for j in range(s_j):
                for k in range(s_k):
                    for l in range(s_l):
                        utils.arr_setitem_4d(arr, _des_bytes(data, pos), i, j, k, l)
        return arr

    # Invalid
    raise ValueError("unsupported <'numpy.ndarray'> dimension [%d]." % ndim)


@cython.cfunc
@cython.inline(True)
def _des_ndarray_unicode(data: str, pos: cython.Py_ssize_t[1]) -> object:
    """(cfunc) Deserialize 'data' at the given position as `<'np.ndarray'>`.

    #### This function is only for ndarray dtype: `"U" (Unicode)`.

    Position updates to the next charactor after deserialization.
    """
    idx: cython.Py_ssize_t = pos[0] + 1  # skip array dtype
    npy_dtype = unpack_uint8(data, idx)  # type: ignore
    ndim: cython.uchar = unpack_uint8(data, idx + 1)  # type: ignore
    pos[0] += 3  # set new position
    arr: np.ndarray
    # 1-dimension
    if ndim == 1:
        # . array shape
        s_i: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        if s_i == 0:
            return np.PyArray_EMPTY(ndim, [s_i], npy_dtype, 0)  # exit: empty
        # . deserialize values
        ch_size: cython.int = dec_encoded_int(data, pos)  # type: ignore
        arr = np.empty(s_i, "U%d" % ch_size)
        values_len: cython.ulonglong = dec_encoded_int(data, pos)  # type: ignore
        idx: cython.Py_ssize_t = pos[0]
        end: cython.Py_ssize_t = idx + values_len
        values_json: str = data[idx:end]
        pos[0] = end  # set new position
        values = iter(_orjson_loads(values_json))
        for i in range(s_i):
            utils.arr_setitem_1d(arr, next(values), i)
        return arr

    # 2-dimension
    if ndim == 2:
        # . array shape
        s_i: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        s_j: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        if s_j == 0:
            return np.PyArray_EMPTY(ndim, [s_i, s_j], npy_dtype, 0)  # exit: empty
        # . deserialize values
        ch_size: cython.int = dec_encoded_int(data, pos)  # type: ignore
        arr = np.empty([s_i, s_j], "U%d" % ch_size)
        values_len: cython.ulonglong = dec_encoded_int(data, pos)  # type: ignore
        idx: cython.Py_ssize_t = pos[0]
        end: cython.Py_ssize_t = idx + values_len
        values_json: str = data[idx:end]
        pos[0] = end  # set new position
        values = iter(_orjson_loads(values_json))
        for i in range(s_i):
            for j in range(s_j):
                utils.arr_setitem_2d(arr, next(values), i, j)
        return arr

    # 3-dimension
    if ndim == 3:
        # . array shape
        s_i: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        s_j: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        s_k: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        if s_k == 0:
            return np.PyArray_EMPTY(ndim, [s_i, s_j, s_k], npy_dtype, 0)  # exit: empty
        # . deserialize values
        ch_size: cython.int = dec_encoded_int(data, pos)  # type: ignore
        arr = np.empty([s_i, s_j, s_k], "U%d" % ch_size)
        values_len: cython.ulonglong = dec_encoded_int(data, pos)  # type: ignore
        idx: cython.Py_ssize_t = pos[0]
        end: cython.Py_ssize_t = idx + values_len
        values_json: str = data[idx:end]
        pos[0] = end  # set new position
        values = iter(_orjson_loads(values_json))
        for i in range(s_i):
            for j in range(s_j):
                for k in range(s_k):
                    utils.arr_setitem_3d(arr, next(values), i, j, k)
        return arr

    # 4-dimension
    if ndim == 4:
        # . array shape
        s_i: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        s_j: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        s_k: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        s_l: cython.Py_ssize_t = dec_encoded_int(data, pos)  # type: ignore
        if s_l == 0:
            return np.PyArray_EMPTY(
                ndim, [s_i, s_j, s_k, s_l], npy_dtype, 0
            )  # exit: empty
        # . deserialize values
        ch_size: cython.int = dec_encoded_int(data, pos)  # type: ignore
        arr = np.empty([s_i, s_j, s_k, s_l], "U%d" % ch_size)
        values_len: cython.ulonglong = dec_encoded_int(data, pos)  # type: ignore
        idx: cython.Py_ssize_t = pos[0]
        end: cython.Py_ssize_t = idx + values_len
        values_json: str = data[idx:end]
        pos[0] = end  # set new position
        values = iter(_orjson_loads(values_json))
        for i in range(s_i):
            for j in range(s_j):
                for k in range(s_k):
                    for l in range(s_l):
                        utils.arr_setitem_4d(arr, next(values), i, j, k, l)
        return arr

    # Invalid
    raise ValueError("unsupported <'numpy.ndarray'> dimension [%d]." % ndim)


# Pandas Types --------------------------------------------------------------------------------
@cython.cfunc
@cython.inline(True)
def _des_pd_timestamp(data: str, pos: cython.Py_ssize_t[1]) -> object:
    """(cfunc) Deserialize 'data' at the given position as `<'pd.Timestamp'>`.

    Position updates to the next byte after deserialization.
    """
    idx: cython.Py_ssize_t = pos[0]
    flag: cython.uchar = unpack_uint8(data, idx + 1)  # type: ignore
    val: cython.longlong = unpack_int64(data, idx + 2)  # type: ignore
    # Without timezone
    if flag == 0:
        pos[0] += 10  # set new position
        return typeref.PD_TIMESTAMP(val)
    # Timezone: approach 1 (tz_name)
    elif flag == 1:
        pos[0] += 10  # set new position
        size: cython.ulonglong = dec_encoded_int(data, pos)  # type: ignore
        idx = pos[0]
        end: cython.ulonglong = idx + size
        tz_name: str = data[idx:end]
        pos[0] = end  # set new position
        return typeref.PD_TIMESTAMP(val, tz=tz_name)
    # Timezone: approach 2 (tz_offset)
    else:
        offset = datetime.timedelta_new(
            unpack_int32(data, idx + 10),  # type: ignore
            unpack_int32(data, idx + 14),  # type: ignore
            unpack_int32(data, idx + 18),  # type: ignore
        )
        pos[0] += 22  # set new position
        tzinfo = datetime.timezone_new(offset, None)
        return typeref.PD_TIMESTAMP(val, tzinfo=tzinfo)


@cython.cfunc
@cython.inline(True)
def _des_pd_timedelta(data: str, pos: cython.Py_ssize_t[1]) -> object:
    """(cfunc) Deserialize 'data' at the given position as `<'pd.Timedelta'>`.

    Position updates to the next byte after deserialization.
    """
    idx: cython.Py_ssize_t = pos[0] + 1  # skip identifier
    val: cython.longlong = unpack_int64(data, idx)  # type: ignore
    pos[0] += 9  # set new position
    return typeref.PD_TIMEDELTA(val)


@cython.cfunc
@cython.inline(True)
def _des_series(data: str, pos: cython.Py_ssize_t[1]) -> object:
    """(cfunc) Deserialize 'data' at the given position as `<'pd.Series'>`.

    Position updates to the next byte after deserialization.
    """
    # Deserialize name
    has_name: cython.uchar = unpack_uint8(data, pos[0] + 1)  # type: ignore
    pos[0] += 2  # skip identifier & has_name
    name = None if has_name == 0 else _des_data(data, pos)

    # Deserialize values
    arr_dtype: cython.uchar = data[pos[0]]
    # . ndarray[object]
    if arr_dtype == identifier.NDARRAY_OBJECT:
        arr = _des_ndarray_object(data, pos)
    # . ndarray[int]
    elif arr_dtype == identifier.NDARRAY_INT:
        arr = _des_ndarray_numeric(data, pos)
    # . ndarray[uint]
    elif arr_dtype == identifier.NDARRAY_UINT:
        arr = _des_ndarray_numeric(data, pos)
    # . ndarray[float]
    elif arr_dtype == identifier.NDARRAY_FLOAT:
        arr = _des_ndarray_numeric(data, pos)
    # . ndarray[bool]
    elif arr_dtype == identifier.NDARRAY_BOOL:
        arr = _des_ndarray_numeric(data, pos)
    # . ndarray[datetime64]
    elif arr_dtype == identifier.NDARRAY_DT64:
        flag: cython.uchar = unpack_uint8(data, pos[0] + 1)  # type: ignore
        pos[0] += 2  # skip arr dtype & flag
        # . Without timezone
        if flag == 0:
            arr = _des_ndarray_nptime(data, pos, True)
        # . Timezone: approach 1 (tz_name)
        elif flag == 1:
            size: cython.ulonglong = dec_encoded_int(data, pos)  # type: ignore
            idx: cython.Py_ssize_t = pos[0]
            end: cython.Py_ssize_t = idx + size
            tz_name: str = data[idx:end]
            pos[0] = end  # set new position
            arr = _des_ndarray_nptime(data, pos, True)
            series = typeref.SERIES(arr, name=name, copy=False)
            return series.dt.tz_localize("UTC").dt.tz_convert(tz_name)  # exit
        # . Timezone: approach 2 (tz_offset)
        else:
            idx: cython.Py_ssize_t = pos[0]
            offset = datetime.timedelta_new(
                unpack_int32(data, idx),  # type: ignore
                unpack_int32(data, idx + 4),  # type: ignore
                unpack_int32(data, idx + 8),  # type: ignore
            )
            pos[0] += 12  # set new position
            tzinfo = datetime.timezone_new(offset, None)
            arr = _des_ndarray_nptime(data, pos, True)
            series = typeref.SERIES(arr, name=name, copy=False)
            return series.dt.tz_localize("UTC").dt.tz_convert(tzinfo)  # exit
    # . ndarray[timedelta64]
    elif arr_dtype == identifier.NDARRAY_TD64:
        arr = _des_ndarray_nptime(data, pos, False)
    # . ndarray[complex]
    elif arr_dtype == identifier.NDARRAY_COMPLEX:
        arr = _des_ndarray_complex(data, pos)
    # . ndarray[bytes]
    elif arr_dtype == identifier.NDARRAY_BYTES:
        arr = _des_ndarray_bytes(data, pos)
    # . ndarray[str]
    elif arr_dtype == identifier.NDARRAY_UNICODE:
        arr = _des_ndarray_unicode(data, pos)
    # . invalid dtype
    else:
        raise ValueError("unsupported <'pandas.Series'> dtype '%s'." % chr(arr_dtype))

    # Generate Series
    return typeref.SERIES(arr, name=name, copy=False)


@cython.cfunc
@cython.inline(True)
def _des_datetime_index(data: str, pos: cython.Py_ssize_t[1]) -> object:
    """(cfunc) Deserialize 'data' at the given position as `<'pd.DatetimeIndex'>`.

    Position updates to the next byte after deserialization.
    """
    # Deserialize name
    has_name: cython.uchar = unpack_uint8(data, pos[0] + 1)  # type: ignore
    pos[0] += 2  # skip identifier & has_name
    name = None if has_name == 0 else _des_data(data, pos)

    # Deserialize values
    flag: cython.uchar = unpack_uint8(data, pos[0])  # type: ignore
    pos[0] += 1  # skip arr dtype & flag
    # . Without timezone
    if flag == 0:
        arr = _des_ndarray_nptime(data, pos, True)
        return typeref.DATETIMEINDEX(arr, name=name, copy=False)
    # . Timezone: approach 1 (tz_name)
    elif flag == 1:
        size: cython.ulonglong = dec_encoded_int(data, pos)  # type: ignore
        idx: cython.Py_ssize_t = pos[0]
        end: cython.Py_ssize_t = idx + size
        tz_name = data[idx:end]
        pos[0] = end  # set new position
        arr = _des_ndarray_nptime(data, pos, True)
        dindex = typeref.DATETIMEINDEX(arr, name=name, copy=False)
        return dindex.tz_localize("UTC").tz_convert(tz_name)
    # . Timezone: approach 2 (tz_offset)
    else:
        idx: cython.Py_ssize_t = pos[0]
        offset = datetime.timedelta_new(
            unpack_int32(data, idx),  # type: ignore
            unpack_int32(data, idx + 4),  # type: ignore
            unpack_int32(data, idx + 8),  # type: ignore
        )
        pos[0] += 12
        tzinfo = datetime.timezone_new(offset, None)
        arr = _des_ndarray_nptime(data, pos, True)
        dindex = typeref.DATETIMEINDEX(arr, name=name, copy=False)
        return dindex.tz_localize("UTC").tz_convert(tzinfo)


@cython.cfunc
@cython.inline(True)
def _des_timedelta_index(data: str, pos: cython.Py_ssize_t[1]) -> object:
    """(cfunc) Deserialize 'data' at the given position as `<'pd.TimedeltaIndex'>`.

    Position updates to the next byte after deserialization.
    """
    # Deserialize name
    has_name: cython.uchar = unpack_uint8(data, pos[0] + 1)  # type: ignore
    pos[0] += 2  # skip identifier & has_name
    name = None if has_name == 0 else _des_data(data, pos)

    # Deserialize values
    arr = _des_ndarray_nptime(data, pos, False)
    return typeref.TIMEDELTAINDEX(arr, name=name, copy=False)


@cython.cfunc
@cython.inline(True)
def _des_dataframe(data: str, pos: cython.Py_ssize_t[1]) -> object:
    """(cfunc) Deserialize 'data' at the given position as `<'pd.DataFrame'>`.

    Position updates to the next byte after deserialization.
    """
    # Deserialize columns name
    has_cols: cython.uchar = unpack_uint8(data, pos[0] + 1)  # type: ignore
    pos[0] += 2  # skip identifier & has_cols
    # . empty w/o columns
    if has_cols == 0:
        return typeref.DATAFRAME()  # exit
    cols: list = _des_list(data, pos)
    has_row: cython.uchar = unpack_uint8(data, pos[0])  # type: ignore
    pos[0] += 1  # skip has_row
    # . empty w/o rows
    if has_row == 0:
        return typeref.DATAFRAME(columns=cols)  # exit

    # Deserialize columns values
    values: dict = {}
    for col in cols:
        arr_dtype: cython.uchar = data[pos[0]]
        # . ndarray[object]
        if arr_dtype == identifier.NDARRAY_OBJECT:
            values[col] = _des_ndarray_object(data, pos)
        # . ndarray[int]
        elif arr_dtype == identifier.NDARRAY_INT:
            values[col] = _des_ndarray_numeric(data, pos)
        # . ndarray[uint]
        elif arr_dtype == identifier.NDARRAY_UINT:
            values[col] = _des_ndarray_numeric(data, pos)
        # . ndarray[float]
        elif arr_dtype == identifier.NDARRAY_FLOAT:
            values[col] = _des_ndarray_numeric(data, pos)
        # . ndarray[bool]
        elif arr_dtype == identifier.NDARRAY_BOOL:
            values[col] = _des_ndarray_numeric(data, pos)
        # . ndarray[datetime64]
        elif arr_dtype == identifier.NDARRAY_DT64:
            flag: cython.uchar = unpack_uint8(data, pos[0] + 1)  # type: ignore
            pos[0] += 2  # skip arr dtype & flag
            # . Without timezone
            if flag == 0:
                values[col] = _des_ndarray_nptime(data, pos, True)
            # . Timezone: approach 1 (tz_name)
            elif flag == 1:
                size: cython.ulonglong = dec_encoded_int(data, pos)  # type: ignore
                idx: cython.Py_ssize_t = pos[0]
                end: cython.Py_ssize_t = idx + size
                tz_name: str = data[idx:end]
                pos[0] = end  # set new position
                arr = _des_ndarray_nptime(data, pos, True)
                dindex = typeref.DATETIMEINDEX(arr, copy=False)
                values[col] = dindex.tz_localize("UTC").tz_convert(tz_name)
            # . Timezone: approach 2 (tz_offset)
            else:
                idx: cython.Py_ssize_t = pos[0]
                offset = datetime.timedelta_new(
                    unpack_int32(data, idx),  # type: ignore
                    unpack_int32(data, idx + 4),  # type: ignore
                    unpack_int32(data, idx + 8),  # type: ignore
                )
                pos[0] += 12  # set new position
                tzinfo = datetime.timezone_new(offset, None)
                arr = _des_ndarray_nptime(data, pos, True)
                dindex = typeref.DATETIMEINDEX(arr, copy=False)
                values[col] = dindex.tz_localize("UTC").tz_convert(tzinfo)
        # . ndarray[timedelta64]
        elif arr_dtype == identifier.NDARRAY_TD64:
            values[col] = _des_ndarray_nptime(data, pos, False)
        # . ndarray[complex]
        elif arr_dtype == identifier.NDARRAY_COMPLEX:
            values[col] = _des_ndarray_complex(data, pos)
        # . ndarray[bytes]
        elif arr_dtype == identifier.NDARRAY_BYTES:
            values[col] = _des_ndarray_bytes(data, pos)
        # . ndarray[str]
        elif arr_dtype == identifier.NDARRAY_UNICODE:
            values[col] = _des_ndarray_unicode(data, pos)
        # . invalid dtype
        else:
            raise ValueError(
                "unsupported <'pandas.DataFrame'> column's dtype '%s'." % chr(arr_dtype)
            )

    # Generate DataFrame
    return typeref.DATAFRAME(values, columns=cols, copy=False)


# Deserialize ---------------------------------------------------------------------------------
@cython.cfunc
@cython.inline(True)
def _des_data(data: str, pos: cython.Py_ssize_t[1]) -> object:
    """(cfunc) Deserialize 'data' at the given position back to an `<'object'>`.

    Position updates to the next byte after deserialization.
    """
    dtype_id: cython.char = str_read(data, pos[0])

    # Basic Types
    # <'str'>
    if dtype_id == identifier.STR:
        return _des_str(data, pos)
    # <'int'>
    if dtype_id == identifier.INT:
        return _des_int(data, pos)
    # <'float'>
    if dtype_id == identifier.FLOAT:
        return _des_float(data, pos)
    # <'bool'>
    if dtype_id == identifier.BOOL:
        return _des_bool(data, pos)
    # <'NoneType'>
    if dtype_id == identifier.NONE:
        return _des_none(data, pos)

    # Date&Time Types
    # <'datetime.datetime'>
    if dtype_id == identifier.DATETIME:
        return _des_datetime(data, pos)
    # <'datetime.date'>
    if dtype_id == identifier.DATE:
        return _des_date(data, pos)
    # <'datetime.time'>
    if dtype_id == identifier.TIME:
        return _des_time(data, pos)
    # <'datetime.timedelta'>
    if dtype_id == identifier.TIMEDELTA:
        return _des_timedelta(data, pos)
    # <'time.struct_time'>
    if dtype_id == identifier.STRUCT_TIME:
        return _des_struct_time(data, pos)
    # <'pandas.Timestamp'>
    if dtype_id == identifier.PD_TIMESTAMP:
        return _des_pd_timestamp(data, pos)
    # <'pandas.Timedelta'>
    if dtype_id == identifier.PD_TIMEDELTA:
        return _des_pd_timedelta(data, pos)
    # <'numpy.datetime64'>
    if dtype_id == identifier.DATETIME64:
        return _des_datetime64(data, pos)
    # <'numpy.timedelta64'>
    if dtype_id == identifier.TIMEDELTA64:
        return _des_timedelta64(data, pos)

    # Numeric Types
    # <'decimal.Decimal'>
    if dtype_id == identifier.DECIMAL:
        return _des_decimal(data, pos)
    # <'complex'>
    if dtype_id == identifier.COMPLEX:
        return _des_complex(data, pos)

    # Bytes Types
    # <'bytes'>
    if dtype_id == identifier.BYTES:
        return _des_bytes(data, pos)
    # <'bytearray'>
    if dtype_id == identifier.BYTEARRAY:
        return _des_bytearray(data, pos)

    # Sequence Types
    # <'list'>
    if dtype_id == identifier.LIST:
        return _des_list(data, pos)
    # <'tuple'>
    if dtype_id == identifier.TUPLE:
        return _des_tuple(data, pos)
    # <'set'>
    if dtype_id == identifier.SET:
        return _des_set(data, pos)
    # <'frozenset'>
    if dtype_id == identifier.FROZENSET:
        return _des_frozenset(data, pos)
    # <'range'>
    if dtype_id == identifier.RANGE:
        return _des_range(data, pos)

    # Mapping Types
    # <'dict'>
    if dtype_id == identifier.DICT:
        return _des_dict(data, pos)

    # NumPy Types
    # <'numpy.ndarray'>
    if dtype_id == identifier.NDARRAY:
        return _des_ndarray(data, pos)

    # Pandas Types
    # <'pandas.Series'>
    if dtype_id == identifier.SERIES:
        return _des_series(data, pos)
    # <'pandas.DataFrame'>
    if dtype_id == identifier.DATAFRAME:
        return _des_dataframe(data, pos)
    # <'pandas.DatetimeIndex'>
    if dtype_id == identifier.DATETIMEINDEX:
        return _des_datetime_index(data, pos)
    # <'pandas.TimedeltaIndex'>
    if dtype_id == identifier.TIMEDELTAINDEX:
        return _des_timedelta_index(data, pos)

    # Invalid 'data'
    raise ValueError("unknown 'data' identifer '%s'." % chr(dtype_id))


@cython.ccall
def deserialize(data: str) -> object:
    """Deserialize the data (str) back to an `<'object'>`."""
    try:
        eof: cython.Py_ssize_t = str_len(data)
        if eof == 0:
            raise ValueError("'data' is empty.")
        pos: cython.Py_ssize_t[1] = [0]
        obj = _des_data(data, pos)
        if pos[0] != eof:
            raise ValueError("improper [pos] %d / [eof] %d." % (pos[0], eof))
        return obj
    except Exception as err:
        raise errors.DeserializeError(
            "<'Serializor'>\nFailed to deserialize from: "
            "%s\n%r\nError: %s" % (type(data), data, err)
        ) from err


########## The following functions are for testing purpose only ##########
def _test_utils() -> None:
    _test_unpack_int8()
    _test_unpack_uint8()
    _test_unpack_int16()
    _test_unpack_uint16()
    _test_unpack_int24()
    _test_unpack_uint24()
    _test_unpack_int32()
    _test_unpack_uint32()
    _test_unpack_int64()
    _test_unpack_uint64()
    _test_dec_encoded_int()


def _test_unpack_int8() -> None:
    import struct

    for val in range(-128, 128):
        b = struct.pack("<b", val).decode("latin1")
        i = unpack_int8(b, 0)  # type: ignore
        assert val == i, f"val {val} vs i {i}"
        print("- pass: %s" % val, end="\r")
    print("Pass Unpack int8".ljust(80))

    del struct


def _test_unpack_uint8() -> None:
    import struct

    for val in range(0, 256):
        b = struct.pack("<B", val).decode("latin1")
        i = unpack_uint8(b, 0)  # type: ignore
        assert val == i, f"val {val} vs i {i}"
        print("- pass: %s" % val, end="\r")
    print("Pass Unpack uint8".ljust(80))

    del struct


def _test_unpack_int16() -> None:
    import struct

    for val in range(-32768, 32768):
        b = struct.pack("<h", val).decode("latin1")
        i = unpack_int16(b, 0)  # type: ignore
        assert val == i, f"val {val} vs i {i}"
        print("- pass: %s" % val, end="\r")
    print("Pass Unpack int16".ljust(80))

    del struct


def _test_unpack_uint16() -> None:
    import struct

    for val in range(0, 65536):
        b = struct.pack("<H", val).decode("latin1")
        i = unpack_uint16(b, 0)  # type: ignore
        assert val == i, f"val {val} vs i {i}"
        print("- pass: %s" % val, end="\r")
    print("Pass Unpack uint16".ljust(80))

    del struct


def _test_unpack_int24() -> None:
    import struct

    for val in (-8388608, -8388607, 0, 8388606, 8388607):
        b = (struct.pack("<i", val)[:3]).decode("latin1")
        i = unpack_int24(b, 0)  # type: ignore
        assert val == i, f"val {val} vs i {i}"
        print("- pass: %s" % val, end="\r")
    print("Pass Unpack int24".ljust(80))

    del struct


def _test_unpack_uint24() -> None:
    import struct

    for val in (0, 1, 16777213, 16777214, 16777215):
        b = (struct.pack("<I", val)[:3]).decode("latin1")
        i = unpack_uint24(b, 0)  # type: ignore
        assert val == i, f"val {val} vs i {i}"
        print("- pass: %s" % val, end="\r")
    print("Pass Unpack uint24".ljust(80))

    del struct


def _test_unpack_int32() -> None:
    import struct

    for val in (-2147483648, -2147483647, 0, 2147483646, 2147483647):
        b = struct.pack("<i", val).decode("latin1")
        i = unpack_int32(b, 0)  # type: ignore
        assert val == i, f"val {val} vs i {i}"
        print("- pass: %s" % val, end="\r")
    print("Pass Unpack int32".ljust(80))

    del struct


def _test_unpack_uint32() -> None:
    import struct

    for val in (0, 1, 4294967293, 4294967294, 4294967295):
        b = struct.pack("<I", val).decode("latin1")
        i = unpack_uint32(b, 0)  # type: ignore
        assert val == i, f"val {val} vs i {i}"
        print("- pass: %s" % val, end="\r")
    print("Pass Unpack uint32".ljust(80))

    del struct


def _test_unpack_int64() -> None:
    import struct

    for val in (
        -9223372036854775808,
        -9223372036854775807,
        0,
        9223372036854775806,
        9223372036854775807,
    ):
        b = struct.pack("<q", val).decode("latin1")
        i = unpack_int64(b, 0)  # type: ignore
        assert val == i, f"val {val} vs i {i}"
        print("- pass: %s" % val, end="\r")
    print("Pass Unpack int64".ljust(80))

    del struct


def _test_unpack_uint64() -> None:
    import struct

    for val in (
        0,
        1,
        18446744073709551613,
        18446744073709551614,
        18446744073709551615,
    ):
        b = struct.pack("<Q", val).decode("latin1")
        i = unpack_uint64(b, 0)  # type: ignore
        assert val == i, f"val {val} vs i {i}"
        print("- pass: %s" % val, end="\r")
    print("Pass Unpack uint64".ljust(80))

    del struct


def _test_dec_encoded_int() -> None:
    for enc, num in [
        ("û", 251),
        ("üü\x00", 252),
        ("üÿÿ", 65535),
        ("ý\x00\x00\x01\x00", 65536),
        ("ýÿÿÿÿ", 4294967295),
        ("þ\x00\x00\x00\x00\x01\x00\x00\x00", 4294967296),
        ("þ\x00\x00\x00\x00\x02\x00\x00\x00", 8589934592),
    ]:
        i = dec_encoded_int(enc, [0])  # type: ignore
        assert num == i, f"num {num} vs i {i} - encode: {enc}"
    print("Pass Dec Encoded Integer".ljust(80))



# Base Exceptions ---------------------------------------------------------------------------------
class SerializorError(Exception):
    """Base class for all exceptions raised by serializor."""


# Serialize Errors --------------------------------------------------------------------------------
class SerializeError(SerializorError):
    """Base class for all exceptions raised by serializor during serialization."""


# Deserialize Errors ------------------------------------------------------------------------------
class DeserializeError(SerializorError):
    """Base class for all exceptions raised by serializor during deserialization."""


# Crypto Errors -----------------------------------------------------------------------------------
class CryptoError(SerializorError):
    """Base class for all exceptions raised by serializor during encryption and decryption."""


class EncryptError(CryptoError, SerializeError):
    """Raised when the object encryption fails."""


class DecryptError(CryptoError, DeserializeError):
    """Raised when the data decryption fails."""
